import collections
import concurrent.futures
import time
import types

from . import packlib
from . import poollib
from . import server_socket
from . import thread


class Server:
    def __init__(self, port, name='Server', workers=1, errors=True, **kwargs):
        self.socket = server_socket.ServerSocket(port, name, **kwargs)
        self.loop = thread.Thread(self._loop, name=f'{name}Loop')
        self.methods = {}
        self.jobs = set()
        self.workers = workers
        self.errors = errors
        self.running = False
        self.pool = poollib.ThreadPool(workers, 'default_pool')
        self.postfn_pool = poollib.ThreadPool(1, 'postfn')
        self.postfn_inp = collections.deque()
        self.postfn_out = collections.deque()
        self.metrics = dict(send=0, recv=0, time=time.time())
        self.pools = [self.pool, self.postfn_pool]

    def bind(self, name, workfn, postfn=None, workers=0):
        assert not self.running
        assert name not in self.methods, name
        if workers:
            pool = poollib.ThreadPool(workers, '{name}_pool')
            self.pools.append(pool)
        else:
            pool = self.pool
        requests = collections.deque()
        available = (workers or self.workers) + 1
        self.methods[name] = types.SimpleNamespace(
            workfn=workfn,
            postfn=postfn,
            pool=pool,
            requests=requests,
            available=available,
        )

    def start(self, block=True):
        assert not self.running
        self.running = True
        self.loop.start()
        if block:
            self.loop.join(timeout=None)

    def close(self, timeout=None, internal=False):
        assert self.running
        self.socket.shutdown()
        self.running = False
        if not internal:
            self.loop.join(timeout)
            self.loop.kill()
        [x.close() for x in self.pools]
        self.socket.close()

    def stats(self):
        now = time.time()
        mets = self.metrics
        self.metrics = dict(send=0, recv=0, time=now)
        dur = now - mets['time']
        stats = {
            'numsend': mets['send'],
            'numrecv': mets['recv'],
            'sendrate': mets['send'] / dur,
            'recvrate': mets['recv'] / dur,
            'requests': sum(len(m.requests) for m in self.methods.values()),
            'jobs': len(self.jobs),
        }
        if any(method.postfn for method in self.methods.values()):
            update = {
                'post_iqueue': len(self.postfn_inp),
                'post_oqueue': len(self.postfn_out),
            }
            stats.update(update)
        return stats

    def __enter__(self):
        self.start(block=False)
        return self

    def __exit__(self, *e):
        self.close()

    def _loop(self):
        methods = list(self.methods.values())
        pending = 0
        while self.running or pending:
            while True:  # Loop syntax used to break on error.
                if not self.running:  # Do not accept further requests.
                    break
                try:
                    addr, data = self.socket.recv(timeout=0.0001)
                except TimeoutError:
                    break
                if len(data) < 8:
                    self._error(addr, bytes(8), 1, 'Message too short')
                    break
                reqnum, data = bytes(data[:8]), data[8:]
                try:
                    strlen = int.from_bytes(data[:8], 'little', signed=False)
                    name = bytes(data[8 : 8 + strlen]).decode('utf-8')
                    data = packlib.unpack(data[8 + strlen :])
                except Exception:
                    self._error(addr, reqnum, 2, 'Could not decode message')
                    break
                if name not in self.methods:
                    self._error(addr, reqnum, 3, f'Unknown method {name}')
                    break
                self.metrics['recv'] += 1
                method = self.methods[name]
                method.requests.append((addr, reqnum, data))
                pending += 1
                break  # We do not actually want to loop.

            for method in methods:
                if method.requests and method.available:
                    method.available -= 1
                    addr, reqnum, data = method.requests.popleft()
                    job = method.pool.submit(method.workfn, *data)
                    job.method = method
                    job.addr = addr
                    job.reqnum = reqnum
                    self.jobs.add(job)
                    if method.postfn:
                        self.postfn_inp.append(job)

            completed, self.jobs = concurrent.futures.wait(
                self.jobs, 0.0001, concurrent.futures.FIRST_COMPLETED
            )
            for job in completed:
                try:
                    data = job.result()
                    if job.method.postfn:
                        data, _ = data
                    data = packlib.pack(data)
                    status = int(0).to_bytes(8, 'little', signed=False)
                    self.socket.send(job.addr, job.reqnum, status, *data)
                    self.metrics['send'] += 1
                except Exception as e:
                    message = f'Error in server method: {e}'
                    self._error(job.addr, job.reqnum, 4, message)
                finally:
                    if not job.method.postfn:
                        job.method.available += 1
                        pending -= 1

            if completed:
                # Call postfns in the order the requests were received.
                while self.postfn_inp and self.postfn_inp[0].done():
                    job = self.postfn_inp.popleft()
                    _, info = job.result()
                    postjob = self.postfn_pool.submit(job.method.postfn, info)
                    postjob.method = job.method
                    self.postfn_out.append(postjob)

            while self.postfn_out and self.postfn_out[0].done():
                postjob = self.postfn_out.popleft()
                postjob.result()  # Check if there was an error.
                postjob.method.available += 1
                pending -= 1

    def _error(self, addr, reqnum, status, message):
        status = status.to_bytes(8, 'little', signed=False)
        data = message.encode('utf-8')
        self.socket.send(addr, reqnum, status, data)
        if self.errors:
            # Wait until the error is delivered to the client and then raise.
            self.close(internal=True)
            raise RuntimeError(message)
        else:
            print(f'Error in server method: {message}')



import math
import struct

import msgpack
import numpy as np

from . import sharray


def pack(data):
    leaves, treedef = tree_flatten(data)
    specs, buffers = [], []
    for value in leaves:
        if value is None:
            specs.append(['none'])
            buffers.append(b'\x00')
        elif isinstance(value, str):
            specs.append(['utf8'])
            buffers.append(value.encode('utf-8'))
        elif isinstance(value, (bytes, bytearray, memoryview)):
            if isinstance(value, memoryview):
                value = value.cast('c')
                assert value.c_contiguous
            specs.append(['bytes'])
            buffers.append(value)
        elif isinstance(value, (np.ndarray, np.generic, int, float)):
            value = np.asarray(value)
            if value.dtype == object:
                raise TypeError(data)
            assert value.data.c_contiguous, (
                'Array is not contiguous in memory. Use '
                + "np.asarray(arr, order='C') before passing the data into pack()."
            )
            specs.append(['array', value.shape, value.dtype.str])
            buffer = value.data.cast('c') if value.size else b'\x00'
            buffers.append(buffer)
        elif isinstance(value, sharray.SharedArray):
            specs.append(['sharray', *value.__getstate__()])
            buffers.append(b'\x00')
        else:
            raise NotImplementedError(type(value))
    buffers = [msgpack.packb(treedef), msgpack.packb(specs), *buffers]
    length = len(buffers).to_bytes(8, 'little', signed=False)
    sizes = struct.pack('<' + ('Q' * len(buffers)), *[len(x) for x in buffers])
    buffers = [length, sizes, *buffers]
    return buffers


def unpack(buffer):
    length = int.from_bytes(buffer[:8], 'little', signed=False)
    buffer = buffer[8:]
    sizes = struct.unpack('<' + ('Q' * length), buffer[: 8 * length])
    buffer = buffer[8 * length :]
    limits = np.cumsum(sizes)
    buffers = [buffer[i:j] for i, j in zip([0, *limits[:-1]], limits)]
    treedef, specs, *buffers = buffers
    treedef = msgpack.unpackb(treedef)
    specs = msgpack.unpackb(specs)
    leaves = []
    for spec, buffer in zip(specs, buffers):
        if spec[0] == 'none':
            assert buffer == b'\x00'
            leaves.append(None)
        elif spec[0] == 'utf8':
            leaves.append(bytes(buffer).decode('utf-8'))
        elif spec[0] == 'bytes':
            leaves.append(buffer)
        elif spec[0] == 'array':
            shape, dtype = spec[1:]
            if not math.prod(shape):
                assert buffer == b'\x00'
                buffer = b''
            leaves.append(np.frombuffer(buffer, dtype).reshape(shape))
        elif spec[0] == 'sharray':
            assert buffer == b'\x00'
            leaves.append(sharray.SharedArray(*spec[1:]))
        else:
            raise NotImplementedError(spec)
    data = tree_unflatten(leaves, treedef)
    return data


def tree_map(fn, *trees, isleaf=None):
    assert trees, 'Provide one or more nested Python structures'
    kw = dict(isleaf=isleaf)
    first = trees[0]
    try:
        assert all(isinstance(x, type(first)) for x in trees)
        if isleaf and isleaf(trees[0]):
            return fn(*trees)
        if isinstance(first, list):
            assert all(len(x) == len(first) for x in trees)
            return [
                tree_map(fn, *[t[i] for t in trees], **kw)
                for i in range(len(first))
            ]
        if isinstance(first, tuple):
            assert all(len(x) == len(first) for x in trees)
            return tuple(
                tree_map(fn, *[t[i] for t in trees], **kw)
                for i in range(len(first))
            )
        if isinstance(first, dict):
            assert all(set(x.keys()) == set(first.keys()) for x in trees)
            return {
                k: tree_map(fn, *[t[k] for t in trees], **kw) for k in first
            }
        if hasattr(first, 'keys') and hasattr(first, 'get'):
            assert all(set(x.keys()) == set(first.keys()) for x in trees)
            return type(first)(
                {k: tree_map(fn, *[t[k] for t in trees], **kw) for k in first}
            )
    except AssertionError:
        raise TypeError('Tree structure differs between arguments')
    return fn(*trees)


def tree_flatten(tree, isleaf=None):
    leaves = []
    tree_map(lambda x: leaves.append(x), tree, isleaf=isleaf)
    structure = tree_map(lambda x: None, tree, isleaf=isleaf)
    return tuple(leaves), structure


def tree_unflatten(leaves, structure):
    leaves = iter(tuple(leaves))
    return tree_map(lambda x: next(leaves), structure)


def tree_equals(xs, ys):
    assert type(xs) == type(ys), (type(xs), type(ys))
    if isinstance(xs, (list, tuple)):
        assert len(xs) == len(ys)
        return all(tree_equals(x, y) for x, y in zip(xs, ys))
    elif isinstance(xs, dict):
        assert xs.keys() == ys.keys()
        return all(tree_equals(xs[k], ys[k]) for k in xs.keys())
    elif isinstance(xs, np.ndarray):
        assert xs.shape == ys.shape, (xs.shape, ys.shape)
        assert xs.dtype == ys.dtype, (xs.dtype, ys.dtype)
        return (xs == ys).all()
    elif isinstance(xs, sharray.SharedArray):
        assert xs.name == ys.name, (xs.name, ys.name)
        info = (xs.array.shape, ys.array.shape)
        assert xs.array.shape == ys.array.shape, info
        assert xs.array.dtype == ys.array.dtype, info
        return (xs == ys).all()
    else:
        return xs == ys



import collections
import dataclasses
import os
import queue
import select
import socket
import sys
import threading
import time

from . import buffers
from . import contextlib
from . import thread


class Disconnected(Exception):
    pass


@dataclasses.dataclass
class Options:
    ipv6: bool = False
    autoconn: bool = True
    max_msg_size: int = 4 * 1024**3
    max_recv_queue: int = 128
    max_send_queue: int = 128
    keepalive_after: float = 10
    keepalive_every: float = 10
    keepalive_fails: int = 10
    logging: bool = True
    logging_color: str = 'yellow'
    connect_wait: float = 0.1
    handshake: str = 'portal_handshake'


class ClientSocket:
    def __init__(self, addr, name='Client', start=True, **kwargs):
        addr = str(addr)
        assert '://' not in addr, addr
        host, port = addr.rsplit(':', 1) if ':' in addr else ('', addr)
        self.name = name
        self.options = Options(**{**contextlib.context.clientkw, **kwargs})
        host = host or ('::1' if self.options.ipv6 else '127.0.0.1')
        self.addr = (host, port)

        self.callbacks_recv = []
        self.callbacks_conn = []
        self.callbacks_disc = []

        self.isconn = threading.Event()
        self.wantconn = threading.Event()
        self.sendq = collections.deque()
        self.recvq = queue.Queue()
        self.get_signal, self.set_signal = os.pipe()

        self.running = True
        self.thread = thread.Thread(self._loop, name=f'{name}Loop')
        start and self.thread.start()

    def start(self):
        self.thread.start()

    @property
    def connected(self):
        return self.isconn.is_set()

    def connect(self, timeout=None):
        if not self.options.autoconn:
            self.wantconn.set()
        return self.isconn.wait(timeout)

    def send(self, *data, timeout=None):
        assert self.running
        if len(self.sendq) > self.options.max_send_queue:
            raise RuntimeError('Too many outgoing messages enqueued')
        self.require_connection(timeout)
        maxsize = self.options.max_msg_size
        self.sendq.append(buffers.SendBuffer(*data, maxsize=maxsize))
        os.write(self.set_signal, bytes(1))

    def recv(self, timeout=None):
        assert self.running
        try:
            if timeout is not None and timeout <= 0.2:
                return self.recvq.get(block=(timeout != 0), timeout=timeout)
            start = time.time()
            while True:
                try:
                    return self.recvq.get(
                        timeout=min(timeout, 0.2) if timeout else 0.2
                    )
                except queue.Empty:
                    timeout = timeout and max(
                        0, timeout - (time.time() - start)
                    )
                    self.require_connection(timeout)
                    if timeout == 0:
                        raise
        except queue.Empty:
            raise TimeoutError

    def close(self, timeout=None):
        self.running = False
        self.thread.join(timeout)
        self.thread.kill()
        os.close(self.get_signal)
        os.close(self.set_signal)

    def require_connection(self, timeout):
        if self.connected:
            return
        if not self.options.autoconn:
            raise Disconnected
        if timeout == 0 or not self.isconn.wait(timeout):
            raise TimeoutError

    def _loop(self):
        recvbuf = buffers.RecvBuffer(maxsize=self.options.max_msg_size)
        sock = None
        poll = select.poll()
        poll.register(self.get_signal, select.POLLIN)
        isconn = False  # Local mirror of self.isconn without the lock.
        writing = False

        while self.running or (self.sendq and isconn):
            if not isconn:
                if not self.options.autoconn and not self.wantconn.wait(
                    timeout=0.2
                ):
                    continue
                sock = self._connect()
                if not sock:
                    break
                poll.register(sock, select.POLLIN)
                self.isconn.set()
                isconn = True
                if not self.options.autoconn:
                    self.wantconn.clear()
                [x() for x in self.callbacks_conn]

            try:
                if not writing:
                    fds = [fd for fd, _ in poll.poll(0.2)]
                    if self.get_signal in fds:
                        writing = True
                        os.read(self.get_signal, 1)

                try:
                    recvbuf.recv(sock)
                    if recvbuf.done():
                        if self.recvq.qsize() > self.options.max_recv_queue:
                            message = 'Too many incoming messages enqueued'
                            raise RuntimeError(message)
                        msg = recvbuf.result()
                        self.recvq.put(msg)
                        [x(msg) for x in self.callbacks_recv]
                        maxsize = self.options.max_msg_size
                        recvbuf = buffers.RecvBuffer(maxsize)
                except BlockingIOError:
                    pass

                if self.sendq:
                    try:
                        self.sendq[0].send(sock)
                        if self.sendq[0].done():
                            self.sendq.popleft()
                            if not self.sendq:
                                writing = False
                    except BlockingIOError:
                        pass
                    except ConnectionResetError:
                        # The server is gone but we may have buffered messages left to
                        # read, so we keep the socket open until recv() fails.
                        pass

            except OSError as e:
                detail = f'{type(e).__name__}'
                detail = f'{detail}: {e}' if str(e) else detail
                self._log(f'Connection to server lost ({detail})')
                self.isconn.clear()
                isconn = False
                poll.unregister(sock)
                sock.close()
                # Clear message queue on disconnect. There is no meaningful concept of
                # sucessful delivery of a message at this level. For example, the
                # server could receive the message but then go down immediately after,
                # without doing anything meaningful with the message. Resending can be
                # done based on response messages at a higher level.
                self.sendq.clear()
                recvbuf = buffers.RecvBuffer(maxsize=self.options.max_msg_size)
                [x() for x in self.callbacks_disc]
                continue

        if sock:
            sock.close()

    def _connect(self):
        self._log(f'Connecting to {self.addr[0]}:{self.addr[1]}')
        once = True
        while self.running:
            # We need to resolve the address regularly.
            host, port = self.addr
            if contextlib.context.resolver:
                host, port = contextlib.context.resolver((host, port))
                assert isinstance(host, str), (host, port)
            port = int(port)
            addr = (host, port, 0, 0) if self.options.ipv6 else (host, port)
            sock = self._create()
            error = None
            try:
                sock.settimeout(10)
                sock.connect(addr)
                self._log('Connection established')
                sock.sendall(self.options.handshake.encode('utf-8'))
                sock.settimeout(0)
                return sock
            except TimeoutError as e:
                error = e
            except ConnectionError as e:
                error = e
            except socket.gaierror as e:
                error = e
            if once:
                self._log(f'Still trying to connect... ({error})')
                once = False
            sock.close()
            time.sleep(self.options.connect_wait)
        return None

    def _create(self):
        if self.options.ipv6:
            sock = socket.socket(socket.AF_INET6, socket.SOCK_STREAM)
            sock.setsockopt(socket.IPPROTO_IPV6, socket.IPV6_V6ONLY, 1)
        else:
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        # sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)  # TODO

        after = self.options.keepalive_after
        every = self.options.keepalive_every
        fails = self.options.keepalive_fails
        if after and every and fails:
            if sys.platform == 'linux':
                sock.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
                sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPIDLE, after)
                sock.setsockopt(
                    socket.IPPROTO_TCP, socket.TCP_KEEPINTVL, every
                )
                sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPCNT, fails)
                if hasattr(socket, 'TCP_USER_TIMEOUT'):  # Linux
                    sock.setsockopt(
                        socket.IPPROTO_TCP,
                        socket.TCP_USER_TIMEOUT,
                        1000 * (after + every * fails),
                    )
            if sys.platform == 'darwin':
                sock.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
                sock.setsockopt(
                    socket.IPPROTO_TCP, socket.TCP_KEEPALIVE, every
                )
            if sys.platform == 'win32':
                sock.ioctl(
                    socket.SIO_KEEPALIVE_VALS, (1, after * 1000, every * 1000)
                )

        return sock

    def _log(self, *args):
        if not self.options.logging:
            return
        contextlib.context.print(
            self.name, *args, color=self.options.logging_color
        )



import multiprocessing as mp

try:
    mp.set_start_method('spawn')
except RuntimeError:
    pass

from .contextlib import context
from .contextlib import initfn
from .contextlib import reset
from .contextlib import setup

from .thread import Thread
from .process import Process
from .futures import Future

from .server_socket import ServerSocket
from .client_socket import ClientSocket
from .client_socket import Disconnected

from .client import Client
from .server import Server
from .batching import BatchServer

from .packlib import pack
from .packlib import unpack
from .packlib import tree_equals

from .sharray import SharedArray

from .utils import free_port
from .utils import proc_alive
from .utils import run

from . import futures



import collections
import dataclasses
import os
import queue
import selectors
import socket

from . import buffers
from . import contextlib
from . import thread


class Connection:
    def __init__(self, sock, addr):
        self.sock = sock
        self.addr = addr
        self.recvbuf = None
        self.handshake = b''
        self.sendbufs = collections.deque()

    def fileno(self):
        return self.sock.fileno()


@dataclasses.dataclass
class Options:
    ipv6: bool = False
    host: str = ''
    max_msg_size: int = 4 * 1024**3
    max_recv_queue: int = 4096
    max_send_queue: int = 4096
    logging: bool = True
    logging_color: str = 'blue'
    handshake: str = 'portal_handshake'


class ServerSocket:
    def __init__(self, port, name='Server', **kwargs):
        if isinstance(port, str):
            assert '://' not in port, port
            port = int(port.rsplit(':', 1)[-1])
        self.name = name
        self.options = Options(**{**contextlib.context.serverkw, **kwargs})
        self.handshake = self.options.handshake.encode('utf-8')
        if self.options.ipv6:
            self.sock = socket.socket(socket.AF_INET6, socket.SOCK_STREAM)
            self.sock.setsockopt(socket.IPPROTO_IPV6, socket.IPV6_V6ONLY, 1)
            self.addr = (self.options.host or '::', port, 0, 0)
        else:
            self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self.addr = (self.options.host or '0.0.0.0', port)
        self.sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        # self.sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)  # TODO
        self._log(f'Binding to {self.addr[0]}:{self.addr[1]}')
        self.sock.bind(self.addr)
        self.sock.setblocking(False)
        self.sock.listen(8192)
        self.get_signal, self.set_signal = os.pipe()
        self.sel = selectors.DefaultSelector()
        self.sel.register(self.get_signal, selectors.EVENT_READ, data='signal')
        self.sel.register(self.sock, selectors.EVENT_READ, data=None)
        self._log(f'Listening at {self.addr[0]}:{self.addr[1]}')
        self.conns = {}
        self.recvq = queue.Queue()  # [(addr, bytes)]
        self.reading = True
        self.running = True
        self.error = None
        self.thread = thread.Thread(self._loop, name=f'{name}Loop', start=True)

    @property
    def connections(self):
        return tuple(self.conns.keys())

    def recv(self, timeout=None):
        if self.error:
            raise self.error
        assert self.running
        try:
            return self.recvq.get(block=(timeout != 0), timeout=timeout)
        except queue.Empty:
            raise TimeoutError

    def send(self, addr, *data):
        if self.error:
            raise self.error
        assert self.running
        if self._numsending() > self.options.max_send_queue:
            raise RuntimeError('Too many outgoing messages enqueued')
        maxsize = self.options.max_msg_size
        try:
            buf = buffers.SendBuffer(*data, maxsize=maxsize)
            self.conns[addr].sendbufs.append(buf)
            os.write(self.set_signal, bytes(1))
        except KeyError:
            self._log('Dropping message to disconnected client')

    def shutdown(self):
        self.reading = False

    def close(self, timeout=None):
        self.running = False
        self.thread.join(timeout)
        [conn.sock.close() for conn in self.conns.values()]
        self.sock.close()
        self.sel.close()
        os.close(self.get_signal)
        os.close(self.set_signal)

    def _loop(self):
        writing = False
        try:
            while self.running or self._numsending():
                for key, mask in self.sel.select(timeout=0.2):
                    if key.data == 'signal':
                        writing = True
                        os.read(self.get_signal, 1)
                    elif key.data is None and self.reading:
                        assert mask & selectors.EVENT_READ
                        self._accept(key.fileobj)
                    elif mask & selectors.EVENT_READ and self.reading:
                        self._recv(key.data)
                if not writing:
                    continue
                pending = [
                    conn for conn in self.conns.values() if conn.sendbufs
                ]
                for conn in pending:
                    try:
                        conn.sendbufs[0].send(conn.sock)
                        if conn.sendbufs[0].done():
                            conn.sendbufs.popleft()
                            if not any(conn.sendbufs for conn in pending):
                                writing = False
                    except BlockingIOError:
                        pass
                    except ConnectionResetError:
                        # The client is gone but we may have buffered messages left
                        # to read, so we keep the socket open until recv() fails.
                        pass
        except Exception as e:
            self.error = e

    def _accept(self, sock):
        sock, addr = sock.accept()
        self._log(f'Accepted connection from {addr[0]}:{addr[1]}')
        sock.setblocking(False)
        conn = Connection(sock, addr)
        self.sel.register(sock, selectors.EVENT_READ, data=conn)
        self.conns[addr] = conn

    def _recv(self, conn):
        if not conn.recvbuf:
            conn.recvbuf = buffers.RecvBuffer(self.options.max_msg_size)
        try:
            if len(conn.handshake) < len(self.handshake):
                self._handshake(conn)
                return
            else:
                conn.recvbuf.recv(conn.sock)
        except OSError as e:
            # For example:
            # - ConnectionResetError
            # - TimeoutError: [Errno 110] Connection timed out
            self._disconnect(conn, e)
            return
        if not conn.recvbuf.done():
            return
        if self.recvq.qsize() > self.options.max_recv_queue:
            raise RuntimeError('Too many incoming messages enqueued')
        self.recvq.put((conn.addr, conn.recvbuf.result()))
        conn.recvbuf = None

    def _disconnect(self, conn, e):
        detail = f'{type(e).__name__}'
        detail = f'{detail}: {e}' if str(e) else detail
        self._log(
            f'Closed connection to {conn.addr[0]}:{conn.addr[1]} ({detail})'
        )
        conn = self.conns.pop(conn.addr)
        if conn.sendbufs:
            count = len(conn.sendbufs)
            conn.sendbufs.clear()
            self._log(f'Dropping {count} messages to disconnected client')
        self.sel.unregister(conn.sock)
        conn.sock.close()

    def _numsending(self):
        return sum(len(x.sendbufs) for x in self.conns.values())

    def _handshake(self, conn):
        assert len(conn.handshake) < len(self.handshake)
        size = len(self.handshake) - len(conn.handshake)
        conn.handshake += conn.sock.recv(size)
        if conn.handshake != self.handshake[: len(conn.handshake)]:
            m = f"Expected handshake '{self.handshake}' got '{conn.handshake}'"
            self._disconnect(conn, ValueError(m))

    def _log(self, *args, **kwargs):
        if not self.options.logging:
            return
        color = self.options.logging_color
        contextlib.context.print(self.name, *args, color=color)
